<?php

namespace Drupal\message_digest;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Database;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Mail\MailManagerInterface;
use Drupal\message_digest\Plugin\Notifier\DigestInterface;
use Drupal\message_notify\Plugin\Notifier\Manager;
use Drupal\user\UserInterface;

/**
 * Digest manager service.
 */
class DigestManager implements DigestManagerInterface {

  /**
   * Database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The digest formatter service.
   *
   * @var \Drupal\message_digest\DigestFormatterInterface
   */
  protected $formatter;

  /**
   * The mail manager service.
   *
   * @var \Drupal\Core\Mail\MailManagerInterface
   *
   * @todo This can be removed if/when the message notify sender service is
   *   used. @see https://www.drupal.org/node/2103013
   */
  protected $mailManager;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * Message notifier plugin manager.
   *
   * @var \Drupal\message_notify\Plugin\Notifier\Manager
   */
  protected $notifierManager;

  /**
   * Constructs the message digest manager service.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   The active database connection.
   * @param \Drupal\message_notify\Plugin\Notifier\Manager $notifier_manager
   *   The message notifier plugin manager service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler service.
   * @param \Drupal\message_digest\DigestFormatterInterface $formatter
   *   The digest formatter service.
   * @param \Drupal\Core\Mail\MailManagerInterface $mail_manager
   *   The mail manager service.
   */
  public function __construct(Connection $connection, Manager $notifier_manager, EntityTypeManagerInterface $entity_type_manager, ModuleHandlerInterface $module_handler, DigestFormatterInterface $formatter, MailManagerInterface $mail_manager) {
    $this->database = $connection;
    $this->entityTypeManager = $entity_type_manager;
    $this->formatter = $formatter;
    $this->mailManager = $mail_manager;
    $this->moduleHandler = $module_handler;
    $this->notifierManager = $notifier_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function cleanupOldMessages() {
    // Removes any message entries from the message_digest table if the
    // corresponding message entity has been deleted.
    // The delete method cannot use a JOIN, so first query to find entries to
    // delete.
    // @see https://www.drupal.org/node/2693899
    $query = $this->database->select('message_digest', 'md');
    $query->leftJoin('message', 'm', 'md.mid = m.mid');
    $query->isNull('m.mid');
    $query->fields('md', ['mid']);
    $query->condition('md.sent', TRUE);
    $result = $query->execute();
    foreach ($result as $row) {
      $this->database->delete('message_digest')
        ->condition('mid', $row->mid)
        ->condition('sent', TRUE)
        ->execute();
    }
  }

  /**
   * Send the actual digest email.
   *
   * @param \Drupal\user\UserInterface $account
   *   Account to deliver the message to.
   * @param string $entity_type
   *   The entity type. Leave empty for global digests.
   * @param string|int $entity_id
   *   The entity ID. Leave empty for global digests.
   * @param string $formatted_message
   *   The formatted message.
   */
  protected function deliverDigest(UserInterface $account, $entity_type, $entity_id, $formatted_message) {
    $params = [
      'body' => $formatted_message,
      'entity_type' => $entity_type,
      'entity_id' => $entity_id,
    ];
    // @todo Use the message_notify sender service.
    // @see https://www.drupal.org/node/2103013
    $this->mailManager->mail('message_digest', 'digest', $account->getEmail(), $account->language(), $params);
  }

  /**
   * {@inheritdoc}
   */
  public function processDigests() {
    foreach ($this->notifierManager->getDefinitions() as $plugin_id => $plugin_definition) {
      $notifier = $this->notifierManager->createInstance($plugin_id, []);
      if (!$notifier instanceof DigestInterface) {
        // Only load the "Digest" notifiers and skip the rest.
        continue;
      }

      // Gather up all the messages into neat little digests and send 'em out.
      // It is up to each digest plugin to manage last sent time, etc.
      // @see \Drupal\message_digest\Plugin\Notifier\DigestBase
      $digests = $notifier->aggregate();
      foreach ($digests as $uid => $entity_types) {
        $account = $this->entityTypeManager->getStorage('user')->load($uid);
        foreach ($entity_types as $entity_type) {
          foreach ($entity_type as $entity_id => $messages) {
            $context = [
              // Reference only, cannot change.
              'entity_type' => $entity_type,
              'entity_id' => $entity_id,
              'view_modes' => array_combine($plugin_definition['viewModes'], $plugin_definition['viewModes']),
              // Set to FALSE to prevent delivery.
              'deliver' => TRUE,
              'messages' => $messages,
            ];
            $this->moduleHandler->alter('message_digest_view_mode', $context, $notifier, $account);
            if ($context['deliver']) {
              $formatted_messages = $this->formatter->format($context['messages'], $context['view_modes'], $account);
              $this->deliverDigest($account, $context['entity_type'], $context['entity_id'], $formatted_messages);
            }
          }
        }
        $notifier->markSent($account);
      }
      $notifier->setLastSent();
    }
  }

}
